import {
  size,
  get,
  reject,
  dropRight,
  map,
  keyBy,
  isEqual,
  last,
} from 'lodash';
import { assign as assignObjects, merge as mergeObjects } from 'lodash/fp';
import { actionTypes } from '../constants';
import { getQueryName } from '../utils/query';
import {
  updateItemInArray,
  createReducer,
  preserveValuesFromState,
  pathToArr,
} from '../utils/reducers';

const {
  DOCUMENT_ADDED,
  GET_SUCCESS,
  LISTENER_RESPONSE,
  CLEAR_DATA,
  DELETE_SUCCESS,
  DOCUMENT_REMOVED,
  DOCUMENT_MODIFIED,
} = actionTypes;

/**
 * Case reducer for modifying a document within a collection or
 * subcollection. When storeAs is being used, subcollections are
 * moved to the level of the storeAs (instead of on their parent doc).
 * @param  {Array} [collectionState=[]] - Redux state of current collection
 * @param  {Object} action - The action that was dispatched
 * @return {Array} State with document modified
 */
function modifyDoc(collectionState, action) {
  return updateItemInArray(collectionState, action.meta.doc, item =>
    // assign is used to prevent the removal of existing subcollections
    assignObjects(item, action.payload.data),
  );
}

function lastDocKey(meta) {
  return meta.subcollections ? last(meta.subcollections).doc : meta.doc;
}
/**
 * Case reducer for adding a document to a collection or subcollection.
 * @param  {Array} [collectionState=[]] - Redux state of current collection
 * @param  {Object} action - The action that was dispatched
 * @return {Array} State with document modified
 */
function addDoc(array = [], action) {
  const { meta, payload } = action;
  return [
    ...array.slice(0, payload.ordered.newIndex),
    { id: lastDocKey(meta), ...payload.data },
    ...array.slice(payload.ordered.newIndex),
  ];
}

/**
 * Case reducer for adding a document to a collection.
 * @param  {Array} collectionState - Redux state of current collection
 * @param  {Object} action - The action that was dispatched
 * @return {Array} State with document modified
 */
function removeDoc(array, action) {
  // Remove doc from collection array
  return reject(array, { id: lastDocKey(action.meta) }); // returns a new array
}

/**
 * Case reducer for writing/updating a whole collection.
 * @param  {Array} collectionState - Redux state of current collection
 * @param  {Object} action - The action that was dispatched
 * @return {Array} State with document modified
 */
function writeCollection(collectionState, action) {
  const { meta, merge = { doc: true, collections: true } } = action;
  if (meta.storeAs) {
    return action.payload.ordered;
  }
  const collectionStateSize = size(collectionState);
  const payloadExists = !!size(action.payload.ordered);

  // Handle doc update (update item in array instead of whole array)
  if (meta.doc && merge.doc && collectionStateSize) {
    // Merge if data already exists
    // Merge with existing ordered array if collection merge enabled
    return modifyDoc(collectionState, action);
  }

  // Merge with existing ordered array (existing as source) if collection merge enabled
  if (collectionStateSize && merge.collections) {
    // Listener response is empty - empty state
    if (!payloadExists) {
      return [];
    }
    // Key existing state collection by id (used to prevent multiple array lookups)
    const existingKeys = collectionState && keyBy(collectionState, 'id');
    // Map new doc data with existing doc data (preserves existing sub-collections)
    return map(action.payload.ordered, newDocObj => {
      const existingDoc = get(existingKeys, newDocObj.id);
      // merge with existing doc if existing doc is not equal
      return !!existingDoc && !isEqual(existingDoc, newDocObj)
        ? { ...existingDoc, ...newDocObj }
        : newDocObj;
    });
  }

  if (meta.doc && collectionStateSize) {
    // Update item in array
    return updateItemInArray(collectionState, meta.doc, item =>
      mergeObjects(item, action.payload.ordered[0]),
    );
  }

  return action.payload.ordered;
}

const actionHandlers = {
  [DOCUMENT_ADDED]: addDoc,
  [DOCUMENT_MODIFIED]: modifyDoc,
  [DOCUMENT_REMOVED]: removeDoc,
  [DELETE_SUCCESS]: removeDoc,
  [LISTENER_RESPONSE]: writeCollection,
  [GET_SUCCESS]: writeCollection,
};

/**
 * Reducer for an ordered collection (stored under path within ordered reducer)
 * which is a map of handlers which are called for different action types.
 * @type {Function}
 */
const orderedCollectionReducer = createReducer(undefined, actionHandlers);

/**
 * Remove the last doc setting from action metadata
 * @param  {String} meta.subcollections - Subcollections which the action
 * associates with
 * @param  {String} meta.doc - Name of Document which the action
 * associates with
 */
function removeLastDocFromMeta(meta) {
  if (!meta.subcollections) {
    return { ...meta, doc: undefined };
  }
  const lastSubcollection = last(meta.subcollections);
  const modifiedSubcollections = [
    ...dropRight(meta.subcollections),
    { ...lastSubcollection, doc: undefined },
  ];
  return { ...meta, subcollections: modifiedSubcollections };
}

/**
 * Get key where ordered collection data is stored within state based
 * on query config.
 * @param  {Object} action - The action that was dispatched
 */
function getStoreUnderKey(action) {
  if (action.meta.oldStoreAs) {
    return action.meta.storeAs || action.meta.collection;
  }
  const queryName = getQueryName(action.meta);

  // Query contains other query params, store under full queryName
  if (queryName.includes('?')) {
    // Remove last doc from query (so that path is of parent collection)
    return getQueryName(removeLastDocFromMeta(action.meta));
  }
  const pathArr = pathToArr(queryName);
  // Return top level key if path is not multiple "/"
  if (pathArr.length <= '1') {
    return action.meta.storeAs || action.meta.collection;
  }

  // Remove last / from path
  return dropRight(pathArr).join('/');
}

/**
 * Reducer for ordered state.
 * @param  {Object} [state={}] - Current ordered redux state
 * @param  {Object} action - The action that was dispatched
 * @param  {String} action.type - Type of action that was dispatched
 * @param  {String} action.meta.collection - Name of Collection which the action
 * associates with
 * @param  {String} action.meta.doc - Name of Document which the action
 * associates with
 * @param  {Array} action.meta.subcollections - Subcollections which the action
 * associates with
 * @param  {String} action.meta.storeAs - Another key within redux store that the
 * action associates with (used for storing data under a path different
 * from its collection/document)
 * @param  {Object} action.payload - Object containing data associated with
 * action
 * @param  {Array} action.payload.ordered - Ordered Array Data associated with
 * action
 * @return {Object} Ordered state after reduction
 */
export default function orderedReducer(state = {}, action) {
  // Return state if action is malformed (i.e. no type, or valid meta)
  if (
    !action.type ||
    !action.meta ||
    (!action.meta.storeAs && !action.meta.collection)
  ) {
    return state;
  }

  // Clear state based on config for CLEAR_DATA action type
  if (action.type === CLEAR_DATA) {
    // support keeping data when logging out - #125
    if (action.preserve && action.preserve.ordered) {
      return preserveValuesFromState(state, action.preserve.ordered, {});
    }
    return {};
  }

  // Return original state if action type is not within actionHandlers
  if (!Object.prototype.hasOwnProperty.call(actionHandlers, action.type)) {
    return state;
  }

  const storeUnderKey = getStoreUnderKey(action);
  const collectionStateSlice = state[storeUnderKey];
  return {
    ...state,
    [storeUnderKey]: orderedCollectionReducer(collectionStateSlice, action),
  };
}
